/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
package org.ow2.orchestra.designer.bpmn.views {
import flash.display.DisplayObject;
import flash.events.Event;
import flash.geom.Point;
import flash.geom.Rectangle;

import mx.binding.utils.ChangeWatcher;
import mx.containers.Canvas;
import mx.events.ChildExistenceChangedEvent;
import mx.events.FlexEvent;

import org.ow2.orchestra.designer.bpmn.model.AbstractElement;
import org.ow2.orchestra.designer.bpmn.model.FlowModel;
import org.ow2.orchestra.designer.bpmn.views.utils.ArrowStyle;
import org.ow2.orchestra.designer.bpmn.views.utils.ArrowUtil;
import org.ow2.orchestra.designer.bpmn.views.utils.GraphicUtils;

public class FlowView extends AbstractElementView {
  [Bindable]
  protected var _flowModel:FlowModel;

  protected var _diagramView:AbstractDiagramView;

  protected var sourceView:AbstractElementView;
  protected var targetView:AbstractElementView;

  protected var style:ArrowStyle;
  protected var _canvas:Canvas;

  [Bindable]
  protected var _sourcePoint:Point = new Point();
  [Bindable]
  protected var _targetPoint:Point = new Point();

  public static const SELECTION_ERROR_MARGIN:Number = 8;

  protected var listenerOnSource:Boolean = false;
  protected var listenerOnTarget:Boolean = false;
  protected var listenerOnProcess:Boolean = false;

  function FlowView(flowModel:FlowModel, diagramView:AbstractDiagramView) {
    this._diagramView = diagramView;
    _flowModel = flowModel;
    watchers.addItem(ChangeWatcher.watch(_flowModel, "sourceElement", modelChanged));
    watchers.addItem(ChangeWatcher.watch(_flowModel, "targetElement", modelChanged));

    // create canvas:
    _canvas = new Canvas();
    this.addChild(canvas);

    //Set the arrow style
    style = new ArrowStyle();
    style.headWidth = 11;
    style.headLength = 17;
    style.color = 0x000000;
    style.headFill = true;
    style.thickness = 0;

    this.addEventListener(FlexEvent.CREATION_COMPLETE, redraw);
    //Initialize properties
    init();

  }


  private function init(event:Event = null):void {
    sourceView = _diagramView.getElementViewByModel(_flowModel.sourceElement);
    targetView = _diagramView.getElementViewByModel(_flowModel.targetElement);
    if (this.listenerOnProcess) {
      _diagramView.removeEventListener(ChildExistenceChangedEvent.CHILD_ADD, this.init);
      this.listenerOnProcess = false;
    }
    if (this.listenerOnSource) {
      sourceView.removeEventListener(FlexEvent.CREATION_COMPLETE, this.init);
      this.listenerOnSource = false;
    }
    if (this.listenerOnTarget) {
      targetView.removeEventListener(FlexEvent.CREATION_COMPLETE, this.init);
      this.listenerOnTarget = false;
    }
    if (sourceView == null || targetView == null) {
      this.listenerOnProcess = true;
      _diagramView.addEventListener(ChildExistenceChangedEvent.CHILD_ADD, this.init, false, 0, true);
    } else if (!sourceView.initialized && !targetView.initialized) {
      this.listenerOnSource = true;
      sourceView.addEventListener(FlexEvent.CREATION_COMPLETE, this.init, false, 0, true);
    } else if (!sourceView.initialized) {
      // Draw the arrow when source view will be created
      this.listenerOnSource = true;
      sourceView.addEventListener(FlexEvent.CREATION_COMPLETE, this.init, false, 0, true);
    } else if (!targetView.initialized) {
      // Draw the arrow when source view will be created
      this.listenerOnTarget = true;
      targetView.addEventListener(FlexEvent.CREATION_COMPLETE, this.init, false, 0, true);
    } else {
      // Draw the arrow
      this.addEventListenersOnView(sourceView);
      this.addEventListenersOnView(targetView);
      redraw(event);
    }
  }

  public override function get model():AbstractElement {
    return _flowModel;
  }

  /**
   * Overrides {@link AbstractElementView#dispose} to remove also {@link FlexEvent.CREATION_COMPLETE} events on source and
   * target views.
   */
  override public function dispose():void {
    super.dispose();
    if (this.listenerOnSource) {
      var sourceView:AbstractElementView = _diagramView.getElementViewByModel(_flowModel.sourceElement);
      sourceView.removeEventListener(FlexEvent.CREATION_COMPLETE, this.init);
    }
    if (this.listenerOnTarget) {
      var targetView:AbstractElementView = _diagramView.getElementViewByModel(_flowModel.targetElement);
      targetView.removeEventListener(FlexEvent.CREATION_COMPLETE, this.init);
    }
    if (sourceView != null) {
      this.removeEventListenersOnView(sourceView);
    }
    if (targetView != null) {
      this.removeEventListenersOnView(targetView);
    }
  }

  protected function modelChanged(event:Event = null):void {
    if (sourceView != null) {
      this.removeEventListenersOnView(sourceView);
    }
    if (targetView != null) {
      this.removeEventListenersOnView(targetView);
    }
    sourceView = _diagramView.getElementViewByModel(_flowModel.sourceElement);
    targetView = _diagramView.getElementViewByModel(_flowModel.targetElement);

    this.addEventListenersOnView(sourceView);
    this.addEventListenersOnView(targetView);
    redraw(event);
  }

  public function redraw(event:Event = null, retries:int = 10):void {
    // get source and target border image
    var sourceImage:DisplayObject = sourceView.borderImage;
    var targetImage:DisplayObject = targetView.borderImage;

    // get image centers
    var sourceCenterPoint:Point = sourceImage.localToGlobal(new Point(sourceImage.width / 2, sourceImage.height / 2));
    var targetCenterPoint:Point = targetImage.localToGlobal(new Point(targetImage.width / 2, targetImage.height / 2));

    // update source and target points
    var globalSourcePoint:Point = GraphicUtils.findIntersection(sourceImage, sourceCenterPoint, targetCenterPoint, _diagramView);
    var globalTargetPoint:Point = GraphicUtils.findIntersection(targetImage, targetCenterPoint, sourceCenterPoint, _diagramView);

    _sourcePoint = this.globalToLocal(globalSourcePoint);
    _targetPoint = this.globalToLocal(globalTargetPoint);

    // draw arrow
    ArrowUtil.drawArrow(canvas.graphics, _sourcePoint, _targetPoint, style);

    // draw condition
    canvas.removeAllChildren();
    if ((globalSourcePoint.equals(sourceCenterPoint) || globalTargetPoint.equals(targetCenterPoint)) && retries > 0) {
      // If intersection not found, try again later.
      // Retries is used to prevent recursive calls.
      this.callLater(redraw, [event, retries - 1]);
    }
  }

  public function set sourcePoint(value:Point):void {
    _sourcePoint = value;
    ArrowUtil.drawArrow(canvas.graphics, _sourcePoint, _targetPoint, style);
  }

  public function set targetPoint(value:Point):void {
    _targetPoint = value;
    ArrowUtil.drawArrow(canvas.graphics, _sourcePoint, _targetPoint, style);
  }

  public function get sourcePoint():Point {
    return _sourcePoint;
  }

  public function get targetPoint():Point {
    return _targetPoint;
  }

  /**
   * Overrides the {@link AbstractElementView#includeInto} to check if target and source points are in the selection
   * rectangle.
   * @param selection A {@link Rectangle} representing the selection
   * @param targetCoordinateSpace The {@link DisplayObject} from which <code>selection</code> coordinates are relative
   * @return Returns true if the selection rectangle is valid to select the object false otherwise
   */
  override public function includeInto(selection:Rectangle, targetCoordinateSpace:DisplayObject):Boolean {
    //TODO: generalize to any kind of elements containing arrows
    if (selection.containsPoint(targetCoordinateSpace.globalToLocal(this.localToGlobal(this.sourcePoint)))
        && selection.containsPoint(targetCoordinateSpace.globalToLocal(this.localToGlobal(this.targetPoint)))) {
      return true;
    }

    return false;
  }

  /**
   * Overrides the {@link AbstractElementView#containsRectangle} to check if any corner of the selection rectangle is
   * separated from the arrow line by less than {@link org.ow2.orchestra.designer.bpmn.views.process.SequenceFlowView#SELECTION_ERROR_MARGIN}
   *
   * @param selection A {@link Rectangle} representing the selection
   * @param targetCoordinateSpace The {@link DisplayObject} from which <code>selection</code> coordinates are relative
   * @return Returns true if the selection rectangle is valid to select the object false otherwise
   */
  override public function containsRectangle(selection:Rectangle, targetCoordinateSpace:DisplayObject):Boolean {
    //TODO: generalize to any kind of elements containing arrows
    var rebasedSourcePoint:Point = targetCoordinateSpace.globalToLocal(this.localToGlobal(this.sourcePoint));
    var rebasedTargetPoint:Point = targetCoordinateSpace.globalToLocal(this.localToGlobal(this.targetPoint));
    var topLeftCornerDistance:Number = GraphicUtils.distanceToSegment(rebasedSourcePoint, rebasedTargetPoint, selection.topLeft);
    var bottomLeftCornerDistance:Number = GraphicUtils.distanceToSegment(rebasedSourcePoint, rebasedTargetPoint, new Point(selection.x,
        selection.y + selection.width));
    var topRightCornerDistance:Number = GraphicUtils.distanceToSegment(rebasedSourcePoint, rebasedTargetPoint, new Point(selection.x
        + selection.height,
        selection.y));
    var bottomRightCornerDistance:Number = GraphicUtils.distanceToSegment(rebasedSourcePoint, rebasedTargetPoint, new Point(selection.x
        + selection.height,
        selection.y + selection.width));

    if (topLeftCornerDistance <= SELECTION_ERROR_MARGIN && bottomLeftCornerDistance <= SELECTION_ERROR_MARGIN && topRightCornerDistance <= SELECTION_ERROR_MARGIN
        && bottomRightCornerDistance <= SELECTION_ERROR_MARGIN) {
      return true;
    }

    return false;
  }

  public function get canvas():Canvas {
    return _canvas;
  }

  private function addEventListenersOnView(view:AbstractElementView):void {
    // Update complete event replace all bellow events ant is fired only when the view is actually updated
    view.addEventListener(FlexEvent.UPDATE_COMPLETE, redraw);
//    view.addEventListener(MoveEvent.MOVE, redraw);
//    view.addEventListener(FlexEvent.CREATION_COMPLETE, redraw);
//    view.addEventListener("widthChanged", redraw);
//    view.addEventListener("heightChanged", redraw);
//    view.addEventListener("xChanged", redraw);
//    view.addEventListener("yChanged", redraw);
  }

  private function removeEventListenersOnView(view:AbstractElementView):void {
    // Update complete event replace all bellow events ant is fired only when the view is actually updated
    view.removeEventListener(FlexEvent.UPDATE_COMPLETE, redraw);
//      view.removeEventListener(MoveEvent.MOVE, redraw);
//      view.removeEventListener(FlexEvent.CREATION_COMPLETE, redraw);
//      view.removeEventListener("widthChanged", redraw);
//      view.removeEventListener("heightChanged", redraw);
//      view.removeEventListener("xChanged", redraw);
//      view.removeEventListener("yChanged", redraw);
  }

}
}
